;; Generated package description from buframe.el  -*- no-byte-compile: t -*-
(define-package "buframe" "0.2.0.20251222.91356" "Buffer-local frames" '((emacs "27.1") (timeout "2.1")) :commit "1859b95f9d550f374c0726e8785a944cbccd8bfb" :authors '(("Al Haji-Ali" . "abdo.haji.ali@gmail.com")) :maintainer '("Al Haji-Ali" . "abdo.haji.ali@gmail.com") :keywords '("buffer" "frames" "convenience") :url "https://github.com/haji-ali/buframe")
